# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Josef Skladanka <jskladan@redhat.com>

from flask import Blueprint, render_template, request, jsonify
import werkzeug.exceptions
from sqlalchemy.orm import exc as orm_exc

from flask_restful import reqparse
from werkzeug.exceptions import HTTPException
from werkzeug.exceptions import BadRequest as JSONBadRequest


from execdb import app, db
from execdb.models.job import Job, BuildStep
from sqlalchemy import desc

import json
import re
from datetime import datetime

from pprint import pformat

main = Blueprint('main', __name__)
BB_URL = app.config['BUILDBOT_FRONTPAGE_URL']
RESULTSDB_URL = app.config['RESULTSDB_FRONTPAGE_URL']

RE_PAGE = re.compile(r"([?&])page=([0-9]+)")
RP = {}
RP['get_jobs'] = reqparse.RequestParser()
RP['get_jobs'].add_argument('page', default=0, type=int, location='args')
RP['get_jobs'].add_argument('limit', default=30, type=int, location='args')


def pagination(q, page, limit):
    # pagination offset
    try:
        page = int(page)
        if page > 0:
            offset = page * limit
            q = q.offset(offset)
    except (TypeError, ValueError):
        pass

    # apply the query limit
    try:
        limit = int(limit)
    except (ValueError, TypeError):
        limit = QUERY_LIMIT

    q = q.limit(limit)
    return q

# TODO: find a better way to do this


def prev_next_urls():
    global RE_PAGE
    try:
        match = RE_PAGE.findall(request.url)
        flag, page = match[0][0], int(match[0][1])
    except IndexError:  # page not found
        if '?' in request.url:
            return None, "%s&page=1" % request.url
        else:
            return None, "%s?page=1" % request.url

    prev = None
    next = None
    prevpage = page - 1
    nextpage = page + 1

    if page > 0:
        prev = RE_PAGE.sub("%spage=%s" % (flag, prevpage), request.url)
    next = RE_PAGE.sub("%spage=%s" % (flag, nextpage), request.url)

    return prev, next


@main.route('/')
@main.route('/index')
@main.route('/jobs', methods=['GET'])
def index():
    try:
        args = RP['get_jobs'].parse_args()
    except JSONBadRequest as error:
        return jsonify({"message": "Bad Request"}), error.code
    except HTTPException as error:
        return jsonify(error.data), error.code

    query = db.session.query(Job).order_by(desc(Job.t_triggered))
    query = pagination(query, args['page'], args['limit'])

    prev, next = prev_next_urls()
    jobs = query.all()

    return render_template('index.html',
                           jobs=jobs,
                           buildbot_url=BB_URL,
                           prev=prev,
                           next=next)


@main.route('/jobs/<uuid>', methods=['GET'])
def show_job(uuid):
    try:
        job = db.session.query(Job).filter(Job.uuid == uuid).one()
    except orm_exc.NoResultFound:
        return 'UUID not found', 404

    job.t_triggered = str(job.t_triggered).split('.')[0]
    return render_template('show_job.html',
                           job=job,
                           buildbot_url=BB_URL,
                           resultsdb_url=RESULTSDB_URL,
                           artifacts_base_url=app.config['ARTIFACTS_BASE_URL'])


@main.route('/jobs/<uuid>/steps', methods=['GET'])
def show_steps(uuid):
    try:
        job = db.session.query(Job).filter(Job.uuid == uuid).one()
    except orm_exc.NoResultFound:
        return 'UUID not found', 404

    steps = dict(
        buildbot_url=BB_URL,
        steps=[],
        job_status=job.current_state,
        job_duration=str(job.build_took),
        )
    name_map = {
        "rm_tmp": "Cleaning-up temporary files",
        "rm_log": "Cleaning-up log files",
        "git":    "Cloning the task repository",
        "runtask": "Running the testsuite",
        "MasterShellCommand": "Creating artifacts directory on master",
        "upload": "Uploading the artifacts to master",
        "MasterShellCommand_1": "Compressing artifacts",
        "MasterShellCommand_2": "Publishing the artifacts",
        }
    for step in job.build_steps:
        if step.status not in ['OK', 'NOT OK', 'INPROGRESS']:
            continue
        s = {}
        s['name'] = step.name
        s['description'] = name_map.get(step.name, step.name)
        s['status'] = step.status
        s['log_url'] = step.log_url
        duration = step.step_took
        minutes = duration.seconds / 60
        seconds = duration.seconds - (minutes*60)
        duration = []
        if minutes:
            duration.append('%s minutes' % minutes)
        duration.append('%s seconds' % seconds)
        s['duration'] = ' '.join(duration)
        steps['steps'].append(s)

    return jsonify(steps)


@main.route('/jobs', methods=['POST'])
def create_job():
    job = Job()

    data = request.json
    job.fedmsg_data = json.dumps(data)

    # FIXME - add validation
    job.taskname = data['taskname']
    job.item = data['item']
    job.item_type = data['item_type']
    job.arch = data['arch']

    db.session.add(job)
    db.session.commit()
    # FIXME - add resultsdb-like serializer
    retval = {
        'id': job.id,
        'uuid': job.uuid,
        't_triggered': job.t_triggered.isoformat()
    }

    return jsonify(retval), 201


def process_bb_status(status_data):
    # grab uuid, build state, properties
    build_properties = status_data['properties']
    uuid = build_properties['uuid'][0]
    build_complete = status_data['complete']

    app.logger.info("Processing data for job {} (complete: {})".format(uuid, build_complete))

    # find job in db
    try:
        job = db.session.query(Job).filter(Job.uuid == uuid).one()
    except orm_exc.NoResultFound:
        app.logger.info("UUID {} not found".format(uuid))
        return 'UUID not found', 400

    # if 'complete' is false, create job
    if not build_complete:
        app.logger.debug("%s -- adding job %s"% (uuid, status_data['number']))

        job.t_build_started = datetime.fromtimestamp(status_data['started_at'])

        job.taskname = build_properties['taskname'][0]
        job.item = build_properties['item'][0]
        job.item_type = build_properties['item_type'][0]
        job.arch = build_properties['arch'][0]
        job.slavename = build_properties['slavename'][0]
        job.link_build_log = '/builders/%s/builds/%s' % (
            status_data['buildrequest']['builderid'],
            status_data['number'])

        db.session.add(job)

        db.session.commit()

    # if 'complete' is true, fill in buildsteps, finish job
    else:
        # add the completed time and state

        job.t_build_ended = datetime.fromtimestamp(status_data['complete_at'])
        # add the build steps
        for step_info in status_data['steps']:

            app.logger.debug("%s -- adding step %s"% (uuid, step_info['name']))
            app.logger.debug("%s -- adding step %s"% (uuid, step_info['name']))
            step = BuildStep(name=step_info['name'])
            step.job = job
            step.started_at = datetime.fromtimestamp(step_info['started_at'])
            step.finished_at = datetime.fromtimestamp(step_info['complete_at'])
            step.data = step_info['state_string']

            # there doesn't seem to be a really reasonable way to tell if a step has failed but
            # this should work well enough for now
            if 'failed' in step_info['state_string']:
                step.status = 'NOT OK'
            else:
                step.status = 'OK'

            db.session.add(step)

        db.session.commit()


@main.route('/buildbot', methods=['POST'])
def bb_push():
    """
    Receives the post-push notifications from buildbot and fills in
    the steps for the job.
    """

    data = request.get_json()
    process_bb_status(data)

    # plain 200 code needs to be returned - otherwise buildbot is
    # endlessly trying to re-send the message.
    # FIXME - add logging for non-200 responses
    return '', 200
